#include<stdio.h>
#include<map>
#include<vector>	// debug
#include<algorithm>	// debug
#include<unordered_map>
typedef unsigned int uint;

//!  
struct State{
	uint id_;					//!< id
	bool fin;					//!<   
	uint in;					//!<  ,   
	std::map<char,State*>out;	//!<   (  )
	typedef std::map<char,State*>::const_iterator it; //!< 

	State():in(0),fin(false),id_(-1) {}
	bool is(char sym) const
	{
		return out.find(sym)!=out.end();
	}
	State *get(char sym) const
	{
		return out.at(sym);
	}
	void set(char sym,State *to)
	{
		out[sym]=to;
	}
	void del(char sym) 
	{
		out.erase(sym);
	}
	bool empty() const
	{
		return out.empty();
	}
	//!  
	bool Eq(const State *s) const
	{
		return (fin==s->fin) && (out==s->out);
	}
	uint Hash() const
	{
		uint hash=7;
		if(fin) hash=45799;
		for(it it=out.begin(); it!=out.end(); it++) {
			hash=hash*13979+it->first*79341+it->second->id_*139+13;
		}
		return hash;
	}
	void DumpDot_(FILE *f) const
	{
		if(fin) fprintf(f,"%d [shape=doublecircle]\n",id_);
		for(it it=out.begin(); it!=out.end(); it++) 
			fprintf(f,"  %d -> %d [label=\"%c\"]\n",id_,it->second->id_,it->first);
	}	
};

//!     
class States{
	uint id_;
protected:
	State *root;		//!<   (  )
	//!  
	State *NewState() 
	{
		State *st=new State;
		st->id_=id_++;
		return st;
	}
	//!  
	State *DupState(State *st)
	{
		State *s=NewState();
		s->fin=st->fin;
		for(State::it it=st->out.begin(); it!=st->out.end(); it++)
			SetLink(s,it->first,it->second);
		return s;
	}
	//!  
	void DelState(State *st) 
	{
		for(State::it it=st->out.begin(); it!=st->out.end(); it++) it->second->in--;
		delete st;
	}
	//!   ,     
	bool DelMarking(State *st) 
	{
		if(st->in==-1) return true;
		st->in=-1;
		for(State::it it=st->out.begin(); it!=st->out.end(); it++) 
			if(DelMarking(it->second)) 
				st->set(it->first,0);
		return false;
	}
	//!  c 
	void DelStates(State *st)
	{
		if(!st) return;
		for(State::it it=st->out.begin(); it!=st->out.end(); it++) DelStates(it->second);
		delete st;
	}
	//!    ,     
	//! \param sym 
	//! \param f   
	//! \param t    (0    )
	void SetLink(State *f,char sym,State *t)
	{
		if(t) {
			if(f->is(sym)) {
				State *i=f->get(sym);
				i->in--;
			}
			f->set(sym,t);
			t->in++;
		} else f->del(sym);
	}
	void GetList_(State *st,std::vector<State*>&list) const
	{
		list.push_back(st);
		for(State::it it=st->out.begin(); it!=st->out.end(); it++) GetList_(it->second,list);
	}
public:
	States():id_(0) { 
		root=NewState();
	}
	~States() {
		DelMarking(root);	//!   (   )
		DelStates(root);	//! 
	} 
	void DumpDot_(const char *filename) const
	{
		char cmd[100];
		std::vector<State*>list;
		FILE *f=fopen(filename,"wt");
		fprintf(f,"digraph {\n");
		fprintf(f,"rankdir=LR;\n");
		fprintf(f,"size=\"4,4\"\n");
		fprintf(f,"node [shape=circle];\n");
		GetList_(root,list);
		std::sort(list.begin(),list.end());
		list.erase(std::unique(list.begin(),list.end()),list.end());
		for(uint n=0; n<list.size(); n++) list[n]->DumpDot_(f);
		fprintf(f,"}\n");
		fclose(f);
		sprintf(cmd,"dot %s -Tpng -O",filename);
		system(cmd);
	}
};

//!     
class Reg{
	std::unordered_map<uint,State*>reg;
public:
	//!   
	State *Get(const State *st) const
	{
		uint hash=st->Hash();
		for(std::unordered_map<uint,State*>::const_iterator it=reg.find(hash); it!=reg.end(); it++) {
			State *s=it->second;
			if(st!=s && st->Eq(s)) return s;
		}
		return 0;
	}
	//!   
	State *Add(State *st)
	{
		uint hash=st->Hash();
		reg.insert(std::make_pair(hash,st));
		return st;
	}
	//!   
	void Del(State *st) 
	{
		uint hash=st->Hash();
		for(std::unordered_map<uint,State*>::iterator it=reg.find(hash); it!=reg.end(); it++) 
			if(it->second==st) {it=reg.erase(it);break;}
	}
};


// -----------------------------------------------------------------------
//!  
class Engine:public States{
	Reg reg;
	
	//!    ,    
	//! \param st    
	//! \param txt  
	State *Split(State *st,const char *txt,bool dup)
	{
		if(st->in>1) dup=true;
		if(dup) st=DupState(st); else reg.Del(st);
		char sym=*txt++;         
		if(!sym || !st->is(sym)) return st;
		SetLink(st,sym,Split(st->get(sym),txt,dup));
		return st;
	}
	//!          
	//! \param st   
	//! \param st1      
	//! \param sym sym      (st -> st1)
	void Merge(State *st,State *st1,char sym)
	{
		State *st2=reg.Get(st1);
		if(st2) {
			SetLink(st,sym,st2);
			reg.Del(st1);
			DelState(st1);
		}
	}
	State *Del(State *st,const char *txt)
	{
		char sym=*txt++;
		reg.Del(st);
		if(sym) {
			if(!st->is(sym)) return st;
			State *s=Del(st->get(sym),txt);
			SetLink(st,sym,s);
			if(s) Merge(st,s,sym);
		}else st->fin=false;
		if(st->empty() && !st->fin) {
			DelState(st);
			return 0;
		}
		return reg.Add(st);
	}
	State *Add(State *st,const char *txt)
	{
		char sym=*txt++;
		reg.Del(st);
		if(sym) {
			State *s;
			if(st->is(sym)) s=Add(st->get(sym),txt);
			else s=Add(NewState(),txt);
			SetLink(st,sym,s);
			Merge(st,s,sym);
		} else st->fin=true;
		return reg.Add(st);
	}
public:
	//!  
	void Add(const char *txt)
	{
		Split(root,txt,false);
		Add(root,txt);
	}
	//!  
	void Del(const char *txt)
	{
		Split(root,txt,false);
		Del(root,txt);
	}
};


void TestCore_()
{
	uint n;
	const char *list1[]={"box","fox","boxes","foxes","nox","fxz","boxil","foxilon","foxil","boxilon",0};
	const char *list2[]={"box",0};
	Engine obj;
	for(n=0; list1[n]; n++) {
		printf("add: %s\n",list1[n]);
		obj.Add(list1[n]);
	}
	for(n=0; list2[n]; n++) {
		printf("add: %s\n",list2[n]);
		obj.Del(list2[n]);
	}
	obj.DumpDot_("core.dot");
}

int main()
{
	TestCore_();
	//system("pause");
	return 0;
}


